// module hu_ftp.cpp
// Simple file transfer via LAN for binary files (no code conversion)
// (not compatible with standard FTP!)
// Server and client are in the same program
// Uses my classes CLIENT_SOCKET, SRVSOCKET and MSGSOCKET
// Compiles with Digital Mars C++ compiler
// Should also compile mit Microsoft Visual C++ V6.0
//
// Private Copyright Dr.E.Huckert 04-2003 

// Protocols:
// G (= get file, from client perspective):
//   client: Gfilename
//   server: file portion with length (length < 0 if error) 
//   client: "OK" or "ERR" as ack on each portion until length of portion <= 0
//
// Versions:
//   2003-05-01 V1.0 first working version
//   2004-01-13 V1.2 fixed wrong file name problem and termination problem

#include <windows.h>
#include <iostream.h>
#include <string>
#include <time.h>
#include <stdlib.h>
#include <fcntl.h>

#include "srvsocket.h"
#include "clsocket.h"
#include "msgsocket.h"

#define SERVER 1
#define CLIENT 0
#define RETRY_MAX 100
 
// --------------- globals ----------------------------
int logging = 0;
// ----------------end of globals ---------------------

// Server: Read a binary file and send it via the socket
// Returns: >= 0 if ok
//          < 0  if error
int send_file(SRVSOCKET *srvsock,
              const char *request)
{
MSGSOCKET *pmsg;
int nb, ret;
int ifd;
static char buf[2048];
  if (logging)
    cout << "Server: in send_file()" << endl;
  //
  // open the requested input file
  ifd = open(request, O_RDONLY + O_BINARY);
  if (logging)
    cout << "Server: in send_file(): file=" << request << " file handle=" 
       << ifd << endl;
  if (ifd < 0)
  {
    cout << "ERROR: cannot open file: " << request << endl;
    pmsg = new MSGSOCKET('E', "Cannot open file", strlen("Cannot open file"));
    ret = srvsock->send(pmsg, 0);
    delete pmsg;
    return -1;
  }
  //
  while (1)
  {
    nb = read(ifd, buf, sizeof(buf));
    pmsg = new MSGSOCKET('P', buf, nb);
    ret = srvsock->send(pmsg, 0);
    delete pmsg;
    if (logging)
      cout << "package sent with " << nb << " bytes" << endl;
    //
    // wait for the acknowledgment
    pmsg = srvsock->recv();
    if (pmsg == NULL)
    {
      cout << "Error: cannot receive acknowledgement for file: " 
           << request << endl;
      return -2;
    }
    if (logging)
    {
      memcpy(buf, pmsg->get_content(), pmsg->get_content_length());
      buf[pmsg->get_content_length()] = 0;
      cout << "ACK received: " << buf << endl;
    }
    if (strncmp("ERR", pmsg->get_content(), 3) == 0)
    {
      delete pmsg;
      cout << "ERROR notified from client" << endl;
      close(ifd);
      return -3;
    }
    delete pmsg;
    if (nb <= 0)
      break;
  }   // end while (1)
  //
  close(ifd);
  cout << "File completeley sent: " << request << endl;
  return 0;
}   // end send_file()

// Do the client work: request a file and receive it
// returns: >=0 if ok
//          < 0 if error
int do_client (const char *hostname,
               int port,
               const char *filename)
{
MSGSOCKET     *pmsg  = NULL;
CLIENT_SOCKET *clsock = NULL;
int           ret    = 0;
int           ofd    = -1;
int           ll, nb = 0;
char          cmd[128];
time_t        t1, t2;
  //
  // open a socket connection to a server 
  // The server must be started
  clsock = new CLIENT_SOCKET();
  clsock->set_logging(logging);
  ret   = clsock->open(hostname, port);
  if (ret < 0)
  {
    cout << "ERROR: client cannot open the socket" << endl;
    return -1;
  }
  if (logging)
    cout << "Client: socket open" << endl;

  while (1)
  {
    if (strlen(filename) == 0)
    {
      // no filename fiven on the command line: ask here
      printf("Dateiname:");
      fflush(stdout);
      gets(cmd);
      if (strlen(cmd) == 0)
        break;
    }
    else
      strcpy(cmd, filename);
    //
    // send a 'G' message with the filename as message body
    pmsg = new MSGSOCKET('G', cmd);
    ret = clsock->send(pmsg, 0);
    delete pmsg;
    if (logging)
      cout << "request(filename) sent:" << cmd << endl;
    if (ret < 0)
    {
      cout << "ERROR: client TCP send error" << endl;
      return -1;
    }
    //
    // open the output file
    ofd = creat(cmd, 0666);
    setmode(ofd, O_BINARY);
    if (logging)
      cout << "creat() file handle:" << ofd << endl;
    if (ofd < 0)
    {
      cout << "ERROR: cannot open local file:" << cmd << endl;
    }
    t1 = time(NULL);
    //
    // receive in a loop while the package length is > 0
    // acknowledge each messge with 'OK'
    while (1)
    {
      // receive the next package from the server 
      pmsg = clsock->recv();
      if (pmsg == NULL)
      {
        cout << "Error: cannot receive acknowledgement for file: " <<
                 filename << endl;
        return -2;
      }
      ll = pmsg->get_content_length();
      if (logging)
        cout << "package received with " << ll << " bytes" << endl; 
      if (ll > 0)
      {
        // write to the file on disk
        int anz;
        anz = ll;
        char *caddr;
        caddr = pmsg->get_content();
        caddr++;   // skip the operation byte
        anz --;    // skip the operation byte
        while (anz > 0)
        {
          nb = write(ofd, caddr, anz);
          anz = anz - nb;
          caddr += nb;
        }
      }
      delete pmsg;
      //
      if (ofd < 0)
        strcpy(cmd, "ERR");
      else
        strcpy(cmd, "OK");
      //
      // acknowledge
      pmsg = new MSGSOCKET('P', cmd);
      ret = clsock->send(pmsg, 0);
      delete pmsg;
      if (logging)
        cout << "ACK sent:" << cmd << endl; 
      if (ret < 0)
      {
        cout << "ERROR: TCP send error" << endl;
        return -3;
      }
      if (ofd < 0)
        break;
      if (ll <= 1)
      {
        if (logging)
          cout << "Last package recognized" << endl;
        break;
      }
    }   // end while (1)
    //
    if (ofd >= 0)
    {
      close(ofd);
      t2 = time(NULL);
      cout << "Time consumed:" << (t2 - t1) << " sec." << endl;
    }
    if (strlen(filename) > 0)
      // filename was specified on the command line
      break;
  }    // end while 1 
  //
  clsock->close();
  delete clsock;
  if (logging)
    cout << "client socket closed" << endl;
  return 0;
}   // end do_client()

// returns: >=0 if ok
//          < 0 if error
int do_server(int port)
{
SRVSOCKET    *psock;
MSGSOCKET    *pmsg;
MSGSOCKET    *pack;
char          answer[512];
int           end_flag = 0;
int           ret      = 0;
int           ret_code = 0;
char          operation = 0;
static char   request[256];
int           retry_count = 0;
  //
  cout << "server role assumed" << endl;
  //
newsocket:
  cout << "Waiting for next request..." << endl;
  //
  psock = new SRVSOCKET();
  psock->setLogging(logging);
  if (psock->open(port) < 0)
  {
    cerr << "Error: cannot open socket" << endl;
    return 1;
  }
  end_flag = 0;
  while (! end_flag)
  {
    pmsg = psock->recv();
    if (pmsg == NULL)
    {
      cerr << "Socket Error?" << endl;
      return 2;
    }
    if (logging)
      cout << "operation=" << pmsg->get_operation() <<
              " content=" << pmsg->get_content() << endl;
    strncpy(request, pmsg->get_content(), pmsg->get_content_length());
    request [pmsg->get_content_length()] = 0;
    operation = pmsg->get_operation();
    delete pmsg;
    //
    if (strcmp("EXIT", request) == 0)
      end_flag = 1;
    if (strcmp("exit", request) == 0)
      end_flag = 1;
    //
    if (operation == 'G')
    {
      ret = send_file(psock, request); 
      end_flag = 1; 
    }
  }   // end while (! end_flag)
  //
  // form an endless loop!!!
  Sleep(200L);
  psock->close();
  delete psock;
  goto newsocket;
  //
  return ret_code;
}   // end do_server()


int main(int argc,
         char *argv[])
{
int       n, ret;  
int       port          = -1;
int       role          = SERVER;
char      hostname[64]  = "localhost";
char      filename[128] = "";
  //
  cout << "hu_ftp V1.2 Copyright Dr.E.Huckert 2003-04" << endl;
  cout << "usage: hu_ftp -port nnn [-host name] [-server|-client] [-v] [-v2]" 
       << "[-i filename]" << endl;

  //
  // Evaluate the command line arguments
  for (n=1; n < argc; n++)
  {
    if (strcmp(argv[n], "-port") == 0)
      port = atoi(argv[n+1]);
    if (strcmp(argv[n], "-host") == 0)
      strcpy(hostname, argv[n+1]);
    if (strcmp(argv[n], "-server") == 0)
      role = SERVER;
    if (strcmp(argv[n], "-client") == 0)
      role = CLIENT;
    if (strcmp(argv[n], "-v") == 0)
      logging = 1;
    if (strcmp(argv[n], "-v2") == 0)
      logging = 2;
    if (strcmp(argv[n], "-i") == 0)
      strcpy(filename, argv[n+1]);
  }
  if (role == SERVER)
  {
    ret = do_server(port);
    if (logging)
      cout << "server operation with success code=" << ret << endl;
  }
  else
  {
    ret = do_client(hostname, port, filename);
    if (logging)
      cout << "client operation with success code=" << ret << endl;
  }
  return 0;
}
